import json
from typing import Any


class MqTaskBody:
    __json_cache: Any | None = None
    __string_cache: str | None = None
    __body: bytes
    __size: int

    def __init__(self, body: bytes, size: int):
        self.__body = body
        self.__size = size

    @property
    def body(self) -> bytes:
        return self.__body

    @property
    def size(self) -> int:
        return self.__size

    def as_string(self):
        if self.__string_cache is None:
            self.__string_cache = self.body.decode()
        return self.__string_cache

    @property
    def text(self) -> str:
        return self.as_string()

    def as_bytes(self):
        return self.body

    def as_json(self):
        if self.__json_cache is None:
            self.__json_cache = json.loads(self.as_string())
        return self.__json_cache

    @property
    def json(self) -> Any:
        return self.as_json()



import logging
from asyncio import AbstractEventLoop
from logging import Logger

import aio_pika
from aio_pika.abc import AbstractRobustConnection

from mqtasks.channel import MqTasksChannel
from mqtasks.message_id_factory import MqTaskMessageIdFactory, MqTaskIdFactory


class MqTasksClient:
    __amqp_connection: str
    __verbose: bool
    __loop: AbstractEventLoop
    __message_id_factory: MqTaskMessageIdFactory
    __task_id_factory: MqTaskIdFactory
    __connection: AbstractRobustConnection | None = None
    __logger: Logger

    def __init__(
            self,
            loop: AbstractEventLoop,
            amqp_connection: str,
            verbose: bool = False,
            logger: Logger | None = None,
            message_id_generator: MqTaskMessageIdFactory | None = None,
            task_id_factory: MqTaskIdFactory | None = None,
    ):
        self.__loop = loop
        self.__amqp_connection = amqp_connection
        self.__verbose = verbose
        self.__message_id_factory = message_id_generator or MqTaskMessageIdFactory()
        self.__task_id_factory = task_id_factory or MqTaskIdFactory()
        self.__logger = logger or logging.getLogger(f"{MqTasksClient.__name__}")

    @property
    def logger(self) -> Logger:
        return self.__logger

    async def close(self):
        if self.__connection is not None:
            await self.__connection.close()

    async def queue(
            self,
            queue_name: str,
    ) -> MqTasksChannel:
        connection: AbstractRobustConnection = self.__connection or await aio_pika.connect_robust(
            self.__amqp_connection,
            loop=self.__loop
        )
        self.__connection = connection

        client = MqTasksChannel(
            connection=connection,
            queue_name=queue_name,
            verbose=self.__verbose,
            loop=self.__loop,
            message_id_factory=self.__message_id_factory,
            logger=self.logger.getChild(queue_name),
            task_id_factory=self.__task_id_factory,
        )

        return client



from enum import StrEnum


class MqResponseStatus(StrEnum):
    SUCCESS: str = "success"
    FAILURE: str = "failure"

    @staticmethod
    def parse(value: str):
        if value == "success":
            return MqResponseStatus.SUCCESS
        if value == "failure":
            return MqResponseStatus.FAILURE
        raise ValueError(f"message {value} is not valid")



from asyncio import AbstractEventLoop
from logging import Logger
from typing import Optional

from aio_pika import Message
from aio_pika.abc import AbstractRobustChannel, AbstractExchange, AbstractQueue
from aiormq.abc import ConfirmationFrameType

from mqtasks.body import MqTaskBody
from mqtasks.headers import MqTaskHeaders
from mqtasks.message_id_factory import MqTaskMessageIdFactory
from mqtasks.response_status import MqResponseStatus
from mqtasks.response_types import MqTaskResponseTypes
from mqtasks.utils import to_json_bytes


class MqTaskContext:
    _channel: AbstractRobustChannel
    _queue: AbstractQueue | None
    _exchange: AbstractExchange | None
    _routing_key: str | None
    _loop: AbstractEventLoop
    _logger: Logger
    _message_id_factory: MqTaskMessageIdFactory

    _message_id: str
    _name: str
    _id: str
    _reply_to: str | None
    _body: MqTaskBody

    def __init__(
            self,
            logger: Logger,
            loop: AbstractEventLoop,
            channel: AbstractRobustChannel,
            queue: AbstractQueue | None,
            exchange: AbstractExchange | None,
            routing_key: str | None,
            message_id_factory: MqTaskMessageIdFactory,
            message_id: str,
            task_name: str,
            task_id: str,
            reply_to: str | None,
            task_body: MqTaskBody
    ):
        self._logger = logger
        self._loop = loop
        self._queue = queue
        self._channel = channel
        self._exchange = exchange
        self._routing_key = routing_key
        self._message_id_factory = message_id_factory
        self._message_id = message_id
        self._name = task_name
        self._id = task_id
        self._reply_to = reply_to
        self._body = task_body

    @property
    def logger(self) -> Logger:
        return self._logger

    @property
    def loop(self) -> AbstractEventLoop:
        return self._loop

    @property
    def message_id(self) -> str:
        return self._message_id

    @property
    def name(self) -> str:
        return self._name

    @property
    def id(self) -> str:
        return self._id

    @property
    def reply_to(self) -> str | None:
        return self._reply_to

    @property
    def body(self) -> MqTaskBody:
        return self._body

    @property
    def message_id_factory(self) -> MqTaskMessageIdFactory:
        return self._message_id_factory

    @property
    def routing_key(self) -> str | None:
        return self._routing_key

    @property
    def exchange(self) -> AbstractExchange | None:
        return self._exchange

    @property
    def is_request(self) -> bool:
        return self._exchange is not None and self._reply_to is not None

    async def publish_data_async(
            self,
            body: bytes | str | object | None = None,
    ) -> Optional[ConfirmationFrameType]:
        if not self.is_request:
            raise Exception(
                f"Task {self.name}, id:{self.id}, message_id:{self.message_id}. It is not a request, so you can't publish data as response")

        data: bytes = to_json_bytes(body)

        return await self._exchange.publish(
            Message(
                headers={
                    MqTaskHeaders.TASK: self._name,
                    MqTaskHeaders.RESPONSE_TO_MESSAGE_ID: self._message_id,
                    MqTaskHeaders.RESPONSE_TYPE: MqTaskResponseTypes.DATA,
                    MqTaskHeaders.RESPONSE_STATUS: MqResponseStatus.SUCCESS
                },
                correlation_id=self._id,
                message_id=self._message_id_factory.new_id(),
                body=data
            ),
            routing_key=self._routing_key,
        )



from aio_pika.abc import AbstractRobustChannel, AbstractIncomingMessage

from mqtasks import MqTaskHeaders


class MqMessageContext:
    __channel: AbstractRobustChannel
    __message: AbstractIncomingMessage
    __wait_task: bool

    def __init__(
            self,
            channel: AbstractRobustChannel,
            message: AbstractIncomingMessage,
            wait_task: bool
    ):
        self.__message = message
        self.__channel = channel
        self.__wait_task = wait_task

    @property
    def wait_task(self):
        return self.__wait_task

    @property
    def channel(self):
        return self.__channel

    @property
    def message(self):
        return self.__message

    @property
    def task_id(self) -> str | None:
        return self.message.correlation_id

    @property
    def reply_to(self) -> str | None:
        return self.message.reply_to

    @property
    def message_id(self) -> str | None:
        return self.message.message_id

    @property
    def task_name(self):
        return self.message.headers[MqTaskHeaders.TASK]



from asyncio import AbstractEventLoop
from logging import Logger
from typing import Callable, Optional
import aio_pika
import aiormq
from aio_pika import ExchangeType
from aio_pika.abc import AbstractRobustConnection, AbstractRobustChannel, AbstractIncomingMessage

from mqtasks.body import MqTaskBody
from mqtasks.headers import MqTaskHeaders
from mqtasks.message import MqTaskMessage
from mqtasks.message_id_factory import MqTaskMessageIdFactory, MqTaskIdFactory
from mqtasks.response_status import MqResponseStatus
from mqtasks.response_types import MqTaskResponseTypes
from mqtasks.utils import to_json_bytes, is_valid_replay_topic, deprecated


class MqTasksChannel:
    __connection: AbstractRobustConnection
    __queue_name: str
    __verbose: bool
    __loop: AbstractEventLoop
    __message_id_factory: MqTaskMessageIdFactory
    __task_id_factory: MqTaskIdFactory
    __logger: Logger

    def __init__(
            self,
            connection: AbstractRobustConnection,
            queue_name: str,
            verbose: bool,
            loop: AbstractEventLoop,
            message_id_factory: MqTaskMessageIdFactory,
            logger: Logger,
            task_id_factory: MqTaskIdFactory,
    ):
        self.__connection = connection
        self.__queue_name = queue_name
        self.__verbose = verbose
        self.__loop = loop
        self.__message_id_factory = message_id_factory
        self.__logger = logger
        self.__task_id_factory = task_id_factory

    @property
    def channel(self) -> AbstractRobustChannel:
        return self.__connection.channel()

    @property
    def logger(self) -> Logger:
        return self.__logger

    async def __request_task_async(
            self,
            task_name: str,
            task_id: str | None = None,
            body: bytes | str | object | None = None,
            message_handler: Callable[[MqTaskMessage], None] | None = None,
            replay_to: str | None = None
    ) -> MqTaskMessage:
        if replay_to is not None:
            if not is_valid_replay_topic(replay_to):
                raise ValueError("replay_to is not valid topic")

        data: bytes = to_json_bytes(body)

        # async with self.__connection:
        routing_key = self.__queue_name
        channel = await self.__connection.channel()

        message_id = self.__message_id_factory.new_id()
        task_id = task_id or self.__task_id_factory.new_id()
        task_replay_to = replay_to if is_valid_replay_topic(replay_to) else f"replay.{task_name}.{task_id}"

        # ------------------------------------------------------------
        # get queue and exchange to request
        task_queue = await channel.get_queue(name=routing_key)
        task_exchange = await channel.get_exchange(name=routing_key)

        # ------------------------------------------------------------
        # declare queue and exchange to response
        response_queue = await channel.declare_queue(name=task_replay_to, durable=True)
        response_exchange = await channel.declare_exchange(name=task_replay_to, type=ExchangeType.DIRECT, durable=True)
        # bind queue to exchange
        await response_queue.bind(response_exchange)

        # we send message to do the task
        await task_exchange.publish(
            aio_pika.Message(
                headers={
                    MqTaskHeaders.TASK: task_name,
                },
                correlation_id=task_id,
                reply_to=task_replay_to,
                message_id=message_id,
                body=data),
            routing_key=routing_key,
        )

        response: MqTaskMessage
        # we will receive response of task
        async with response_queue.iterator() as queue_iter:
            message: AbstractIncomingMessage
            async for message in queue_iter:
                async with message.process():
                    if message.headers[MqTaskHeaders.RESPONSE_TYPE] == MqTaskResponseTypes.DATA:
                        if message_handler is not None:
                            message_handler(
                                MqTaskMessage(
                                    logger=self.logger,
                                    loop=self.__loop,
                                    message_id=message.message_id,
                                    task_name=message.headers[MqTaskHeaders.TASK],
                                    task_id=message.correlation_id,
                                    task_body=MqTaskBody(
                                        body=message.body, size=message.body_size
                                    ),
                                    status=MqResponseStatus.parse(message.headers[MqTaskHeaders.RESPONSE_STATUS]),
                                ))
                    elif message.headers[MqTaskHeaders.RESPONSE_TYPE] == MqTaskResponseTypes.RESPONSE:
                        response = MqTaskMessage(
                            logger=self.logger,
                            loop=self.__loop,
                            message_id=message.message_id,
                            task_name=message.headers[MqTaskHeaders.TASK],
                            task_id=message.correlation_id,
                            task_body=MqTaskBody(
                                body=message.body, size=message.body_size
                            ),
                            status=MqResponseStatus.parse(message.headers[MqTaskHeaders.RESPONSE_STATUS]),
                        )
                        break
                    else:
                        raise ValueError(f"response type={message.headers[MqTaskHeaders.RESPONSE_TYPE]} is not valid")

        await response_queue.unbind(response_exchange)
        await response_queue.delete()
        await response_exchange.delete()
        await channel.close()

        return response

    async def exec_task_async(
            self,
            task_name: str,
            task_id: str | None = None,
            body: bytes | str | object | None = None
    ) -> Optional[aiormq.abc.ConfirmationFrameType]:
        data: bytes = to_json_bytes(body)

        routing_key = self.__queue_name
        channel = await self.__connection.channel()

        message_id = self.__message_id_factory.new_id()
        task_id = task_id or self.__task_id_factory.new_id()

        task_exchange = await channel.get_exchange(name=routing_key)

        return await task_exchange.publish(
            aio_pika.Message(
                headers={
                    MqTaskHeaders.TASK: task_name,
                },
                correlation_id=task_id,
                message_id=message_id,
                body=data),
            routing_key=routing_key,
        )

    @deprecated("use request_task_async instead")
    async def run_task_async(
            self,
            task_name: str,
            task_id: str | None = None,
            body: bytes | str | object | None = None,
            message_handler: Callable[[MqTaskMessage], None] | None = None,
            replay_to: str | None = None
    ) -> MqTaskMessage:
        return await self.__request_task_async(
            task_name=task_name,
            task_id=task_id,
            body=body,
            message_handler=message_handler,
            replay_to=replay_to
        )

    async def request_task_async(
            self,
            task_name: str,
            task_id: str | None = None,
            body: bytes | str | object | None = None,
            message_handler: Callable[[MqTaskMessage], None] | None = None,
            replay_to: str | None = None
    ) -> MqTaskMessage:
        return await self.__request_task_async(
            task_name=task_name,
            task_id=task_id,
            body=body,
            message_handler=message_handler,
            replay_to=replay_to
        )



class MqTaskHeaders:
    TASK: str = "task"
    RESPONSE_TO_MESSAGE_ID: str = "response_to_message_id"
    RESPONSE_TYPE: str = "response_type"
    RESPONSE_STATUS: str = "response_status"
    RESPONSE_ERROR_MESSAGE: str = "response_error_message"



import asyncio
import logging

from example.example_config import CONNECTION, QUEUE_NANE_REQUEST_01, QUEUE_NANE_REQUEST_02, QUEUE_NANE_PUBLISH_01
from mqtasks import MqTasksClient

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger("CLIENT")
logger.setLevel(logging.DEBUG)

loop = asyncio.get_event_loop()
client = MqTasksClient(
    loop=loop,
    amqp_connection=CONNECTION,
    logger=logger,
    verbose=True
)


async def exec_task_async(task_name: str, queue: str, body: str | object | None = None) -> None:
    channel = await client.queue(queue)
    return await channel.exec_task_async(
        task_name=task_name,
        body=body,
    )

async def request_task_async(task_name: str, queue: str, body: str | object | None = None) -> None:
    channel = await client.queue(queue)
    response = await channel.request_task_async(
        task_name=task_name,
        body=body,
        message_handler=lambda msg: print(msg)
    )
    print(response)


# ==============================================================================


loop.run_until_complete(request_task_async(task_name="hello_sync", queue=QUEUE_NANE_REQUEST_01, body={"message": "hello sync task1"}))
loop.run_until_complete(request_task_async(task_name="hello_async", queue=QUEUE_NANE_REQUEST_02, body={"message": "hello async task2"}))
loop.run_until_complete(asyncio.sleep(3))
loop.run_until_complete(request_task_async(task_name="hello_sync", queue=QUEUE_NANE_REQUEST_01, body={"message": "hello sync task3"}))
loop.run_until_complete(request_task_async(task_name="hello_async", queue=QUEUE_NANE_REQUEST_02, body={"message": "Привіт, перевірка килилиці"}))

loop.run_until_complete(
    request_task_async(task_name="data_async", queue=QUEUE_NANE_REQUEST_01, body={"message": "async progress task"}))

loop.run_until_complete(exec_task_async(task_name="task_async", queue=QUEUE_NANE_PUBLISH_01, body={"message": "publish_async"}))

loop.run_until_complete(request_task_async(task_name="no_task", queue=QUEUE_NANE_REQUEST_02, body={"message": "no_task"}))

loop.run_until_complete(client.close())

loop.close()



import asyncio
from asyncio import AbstractEventLoop

from .body import MqTaskBody
from .channel import MqTasksChannel
from .client import MqTasksClient
from .context import MqTaskContext
from .headers import MqTaskHeaders
from .message import MqTaskMessage
from .message_id_factory import MqTaskMessageIdFactory
from .mqtasks import MqTasks
from .register import MqTaskRegister
from .response_types import MqTaskResponseTypes


def run_mqtasks(
        tasks: list[MqTasks],
        loop: AbstractEventLoop | None = None
):
    loop = loop if loop is not None else asyncio.get_event_loop()
    for task in tasks:
        loop.create_task(task.run_async(loop))
    loop.run_forever()



import inspect
import logging
from typing import Any, Dict

from aio_pika import Message
from pamqp.common import FieldValue

from mqtasks.context import MqTaskContext
from mqtasks.headers import MqTaskHeaders
from mqtasks.response_status import MqResponseStatus
from mqtasks.response_types import MqTaskResponseTypes
from mqtasks.utils import to_json_bytes


class MqTaskRegister:
    def __init__(
            self,
            name: str,
            func: Any,
            log_body: bool = True
    ):
        self.name = name
        self.func = func
        self.log_body = log_body

    async def invoke_async(self, ctx: MqTaskContext):
        if ctx.logger.isEnabledFor(logging.DEBUG):
            ctx.logger.debug("______________________________________________")
            ctx.logger.debug(f"invoke begin task:{ctx.name} with_id:{ctx.id}")

        has_exception: bool = False
        exception_msq: str | None = None
        func_result: Any | None = None
        try:
            if inspect.iscoroutinefunction(self.func):
                func_result = await self.func(ctx)
            else:
                func_result = self.func(ctx)
        except Exception as e:
            has_exception = True
            exception_msq = str(e)
            ctx.logger.exception(e)
        except Any as e:
            has_exception = True
            exception_msq = str(e)
            ctx.logger.error(e)

        if has_exception:
            func_result = exception_msq

        task_status = MqResponseStatus.FAILURE if has_exception else MqResponseStatus.SUCCESS

        if not ctx.is_request and func_result is not None:
            ctx.logger.error(
                f"task:{ctx.name} with_id:{ctx.id}, status:{task_status}, return value must be None, because it is not a request, current:{str(func_result)}")

        if ctx.is_request:
            data_result: bytes | None = to_json_bytes(func_result)

            if ctx.exchange is not None:
                headers: Dict[str, FieldValue] = {
                    MqTaskHeaders.TASK: ctx.name,
                    MqTaskHeaders.RESPONSE_TO_MESSAGE_ID: ctx.message_id,
                    MqTaskHeaders.RESPONSE_TYPE: MqTaskResponseTypes.RESPONSE,
                    MqTaskHeaders.RESPONSE_STATUS: task_status
                }
                if has_exception:
                    headers[MqTaskHeaders.RESPONSE_ERROR_MESSAGE] = exception_msq

                await ctx.exchange.publish(
                    Message(
                        headers=headers,
                        correlation_id=ctx.id,
                        message_id=ctx.message_id_factory.new_id(),
                        body=data_result or bytes()
                    ),
                    routing_key=ctx.routing_key,
                )

        if ctx.logger.isEnabledFor(logging.DEBUG):
            ctx.logger.debug(
                f"invoke end task:{ctx.name}, with_id:{ctx.id}, status:{task_status.value}, result:{func_result}")
            ctx.logger.debug("--------------------------------------------")



import asyncio
import logging
from mqtasks import run_mqtasks

from example.example_config import CONNECTION, QUEUE_NANE_REQUEST_01, QUEUE_NANE_REQUEST_02, QUEUE_NANE_PUBLISH_01
from mqtasks import MqTasks, MqTaskContext

logging.basicConfig(level=logging.DEBUG)
# CREATE NEW MESSAGE QUEUE TASK HANDLER
tasks = MqTasks(
    amqp_connection=CONNECTION,
    queue_name=QUEUE_NANE_REQUEST_01,
    logger=logging.getLogger(QUEUE_NANE_REQUEST_01),
    logging_level=logging.DEBUG
)

lists = MqTasks(
    amqp_connection=CONNECTION,
    queue_name=QUEUE_NANE_REQUEST_02,
    logger=logging.getLogger(QUEUE_NANE_REQUEST_02),
    logging_level=logging.DEBUG,
)

publish = MqTasks(
    amqp_connection=CONNECTION,
    queue_name=QUEUE_NANE_PUBLISH_01,
    logger=logging.getLogger(QUEUE_NANE_PUBLISH_01),
    logging_level=logging.DEBUG,
)


# DECLARE THE SYNC REQUEST TASK
@tasks.task(name="hello_sync")
def hello_sync(ctx: MqTaskContext):
    print(f"{'request' if ctx.is_request else 'task'} mid:{ctx.message_id} name:{ctx.name} id:{ctx.id} reply_to:{ctx.reply_to} body:{ctx.body.as_json()}")

    # test exception
    raise Exception("hello_sync")

    return {
        "message": "Hello world too!!! :)"
    }


# DECLARE THE ASYNC REQUEST TASK
@lists.task(name="hello_async")
async def hello_async(ctx: MqTaskContext):
    print(f"{'request' if ctx.is_request else 'task'} mid:{ctx.message_id} name:{ctx.name} id:{ctx.id} reply_to:{ctx.reply_to} body:{ctx.body.as_json()}")
    # sleep
    await asyncio.sleep(1)
    return {
        "message": "Hello world too!!! :)"
    }


# DECLARE THE ASYNC REQUEST TASK
@tasks.task(name="data_async")
async def data_async(ctx: MqTaskContext):
    print(f"{'request' if ctx.is_request else 'task'} mid:{ctx.message_id} name:{ctx.name} id:{ctx.id} reply_to:{ctx.reply_to} body:{ctx.body.as_json()}")
    # sleep
    await asyncio.sleep(1)
    await ctx.publish_data_async(body={"progress": 0.2})
    # await asyncio.sleep(5)
    await ctx.publish_data_async(body={"progress": 0.5})
    # await asyncio.sleep(5)
    await ctx.publish_data_async(body={"progress": 0.7})
    return {
        "progress": 1
    }

# DECLARE THE ASYNC TASK
@publish.task(name="task_async", log_body=False)
async def task_async(ctx: MqTaskContext):
    await asyncio.sleep(0)
    print(f"{'request' if ctx.is_request else 'task'} mid:{ctx.message_id} name:{ctx.name} id:{ctx.id} reply_to:{ctx.reply_to} body:{ctx.body.as_json()}")
    return {}

# NEED TO RUN
run_mqtasks(tasks=[
    tasks, lists, publish
])



import asyncio
import logging
from asyncio import AbstractEventLoop
from logging import Logger
from typing import Callable, Any, Coroutine

import aio_pika
import aio_pika.abc
from aio_pika.abc import (
    AbstractIncomingMessage,
    ExchangeType,
    AbstractExchange,
    AbstractQueue,
    AbstractRobustConnection,
    ConsumerTag,
    AbstractRobustChannel,
    AbstractRobustQueue, AbstractRobustExchange,
)

from mqtasks.body import MqTaskBody
from mqtasks.context import MqTaskContext
from mqtasks.message_id_factory import MqTaskMessageIdFactory
from mqtasks.mqtask_strategy import MqTasksConsumeStrategy
from mqtasks.mqtasks_message_context import MqMessageContext
from mqtasks.register import MqTaskRegister


class MqTasks:
    __tasks: dict[str, MqTaskRegister] = dict()
    __amqp_connection: str
    __queue_name: str
    __loop: AbstractEventLoop
    __prefetch_count: int
    __message_id_factory: MqTaskMessageIdFactory
    __logging_level: int
    __wait_invoke_task: bool = False
    __middlewares: list[Callable[[MqMessageContext, Callable[[MqMessageContext], Any]], Any]] = []
    __message_queue: list[AbstractIncomingMessage] = []
    __consumer_tag: ConsumerTag | None = None
    __middleware_chain: Callable[[Any], Coroutine[Any, Any, None]] | None = None
    __is_terminated: bool = False
    __consume_strategy: MqTasksConsumeStrategy = MqTasksConsumeStrategy.QUEUE

    __queue: AbstractRobustQueue
    __exchange: AbstractRobustExchange
    __channel: AbstractRobustChannel

    def __init__(
            self,
            amqp_connection: str,
            queue_name: str,
            prefetch_count: int = 1,
            logger: Logger | None = None,
            message_id_factory: MqTaskMessageIdFactory | None = None,
            logging_level: int = logging.INFO,
            wait_invoke_task: bool = False,
            consume_strategy: MqTasksConsumeStrategy = MqTasksConsumeStrategy.QUEUE
    ):
        self.__amqp_connection = amqp_connection
        self.__queue_name = queue_name
        self.__prefetch_count = prefetch_count
        self.__logger = logger
        self.__message_id_factory = message_id_factory or MqTaskMessageIdFactory()
        self.__logging_level = logging_level
        self.__wait_invoke_task = wait_invoke_task
        self.__middlewares = []
        self.__message_queue = []
        self.__middleware_chain = None
        self.__consume_strategy = consume_strategy

        if self.__logger is None:
            self.__logger = logging.getLogger(f"{MqTasks.__name__}.{queue_name}")
            self.__logger.setLevel(logging_level)
        else:
            self.__logger = logger.getChild(f"{MqTasks.__name__}.{queue_name}")

    @property
    def __if_log(self):
        return self.__logger.isEnabledFor(self.__logging_level)

    def __log(self, msg):
        self.__logger.log(self.__logging_level, msg)

    def __log_line(self):
        self.__logger.log(self.__logging_level, "------------------------------")

    async def __process_message(
            self,
            context: MqMessageContext
    ):
        channel: AbstractRobustChannel = context.channel
        message: AbstractIncomingMessage = context.message
        wait_task: bool = context.wait_task

        task_name = context.task_name

        register: MqTaskRegister
        if task_name in self.__tasks:
            register = self.__tasks[task_name]
        else:
            def raise_exception(ctx: MqTaskContext):
                raise Exception(
                    f"task:'{task_name}' is not registered, task_id:'{ctx.id}', reply_to:'{ctx.reply_to}'")

            register = MqTaskRegister(name=task_name, func=raise_exception, log_body=False)

        task_id: str | None = context.task_id
        reply_to: str | None = context.reply_to
        message_id: str | None = context.message_id

        reply_to_exchange: AbstractExchange | None = None
        reply_to_queue: AbstractQueue | None = None
        if reply_to is not None and reply_to != "":
            reply_to_exchange = await channel.declare_exchange(
                name=reply_to,
                durable=True,
                type=ExchangeType.DIRECT,
                auto_delete=False
            )
            reply_to_queue = await channel.declare_queue(
                name=reply_to,
                durable=True
            )

        if self.__if_log:
            self.__log_line()
            self.__log(f"task: {task_name}")
            self.__log(f"task_id: {task_id}")
            self.__log(f"reply_to: {reply_to}")
            self.__log(f"message_id: {message_id}")
            self.__log(f"content_encoding: {message.content_encoding}")
            self.__log(f"content_type: {message.content_type}")
            self.__log(f"delivery_mode: {message.delivery_mode}")
            self.__log(f"expiration: {message.expiration}")
            self.__log(f"priority: {message.priority}")
            self.__log(f"headers: {str(message.headers)}")
            if register.log_body:
                self.__log("body:")
                self.__log(message.body.decode('unicode_escape'))
            else:
                self.__log("body: log disabled")
            self.__log_line()

        invoke_task = self.loop.create_task(register.invoke_async(
            MqTaskContext(
                logger=self.__logger,
                loop=self.__loop,
                channel=channel,
                queue=reply_to_queue,
                exchange=reply_to_exchange,
                routing_key=reply_to,
                message_id_factory=self.__message_id_factory,
                message_id=message_id,
                task_name=task_name,
                task_id=task_id,
                reply_to=reply_to,
                task_body=MqTaskBody(
                    body=message.body, size=message.body_size
                )),
        ))

        if self.__wait_invoke_task or wait_task:
            await invoke_task

    def __build_middleware_chain(
            self,
            middlewares: list[Callable[[MqMessageContext, Callable[[MqMessageContext], Any]], Any]],
            final_handler: Callable[[MqMessageContext], Any]
    ):
        async def next_middleware(message, index):
            if index < len(middlewares):
                await middlewares[index](message, lambda msg: next_middleware(msg, index + 1))
            else:
                await final_handler(message)

        return lambda message: next_middleware(message, 0)

    async def __consume(self, message: AbstractIncomingMessage):
        if self.__if_log:
            self.__log(f"Consuming, message: {message}")

        if self.__consume_strategy is MqTasksConsumeStrategy.QUEUE:
            await self.__stop_consume()

        async with message.process():
            self.__message_queue.append(message)
            await self.__process_message(MqMessageContext(
                channel=self.__channel,
                message=message,
                wait_task=False
            ))
            self.__message_queue.remove(message)

        if self.__consume_strategy is MqTasksConsumeStrategy.QUEUE:
            await self.__start_consume()

    async def __start_consume(self):
        if self.__if_log:
            self.__log(f"Starting consuming")

        try:
            queue: AbstractRobustQueue = self.__queue
            self.__consumer_tag = await queue.consume(callback=self.__consume, no_ack=False)
        except Exception as e:
            if self.__if_log:
                self.__log(f"Exception:{e}")

    async def __stop_consume(self):
        if self.__if_log:
            self.__log(f"Stopping consuming")

        tag: ConsumerTag | None = self.__consumer_tag
        self.__consumer_tag = None
        await self.__queue.cancel(tag)

    async def __start_connection(self, loop: AbstractEventLoop | None):
        if self.__if_log:
            self.__log(f"aio_pika.connect_robust->begin connection:{self.__amqp_connection}")
        connection: AbstractRobustConnection = await aio_pika.connect_robust(
            self.__amqp_connection,
            loop=loop
        )
        if self.__if_log:
            self.__log(f"aio_pika.connect_robust->end connection:{self.__amqp_connection}")
            self.__log_line()

        async with connection:

            if self.__if_log:
                self.__log("connection.channel()->begin")
            channel: AbstractRobustChannel = await connection.channel()
            self.__channel = channel
            if self.__if_log:
                self.__log("connection.channel()->end")
                self.__log_line()

            await channel.set_qos(prefetch_count=self.__prefetch_count)

            if self.__if_log:
                self.__log(f"channel.declare_exchange->begin exchange:{self.__queue_name}")
            exchange: AbstractRobustExchange = await channel.declare_exchange(
                name=self.__queue_name,
                type=ExchangeType.DIRECT,
                durable=True,
                auto_delete=False
            )
            self.__exchange = exchange
            if self.__if_log:
                self.__log(f"channel.declare_exchange->end exchange:{self.__queue_name}")
                self.__log_line()

            if self.__if_log:
                self.__log(f"channel.declare_queue->begin queue:{self.__queue_name}")
            queue: AbstractRobustQueue = await channel.declare_queue(
                self.__queue_name,
                auto_delete=False,
                durable=True
            )
            self.__queue = queue
            if self.__if_log:
                self.__log(f"channel.declare_queue->end queue:{self.__queue_name}")
                self.__log_line()

            if self.__if_log:
                self.__log(f"queue.bind->begin queue:{self.__queue_name}")
            await queue.bind(exchange, self.__queue_name)
            if self.__if_log:
                self.__log(f"queue.bind->end queue:{self.__queue_name}")
                self.__log_line()

            if len(self.__message_queue) != 0:
                if self.__if_log:
                    self.__log(f"restore messages from queue, count:{len(self.__message_queue)}")
                while len(self.__message_queue) != 0:
                    msg: AbstractIncomingMessage = self.__message_queue[0]
                    await self.__process_message(
                        context=MqMessageContext(
                            channel=channel,
                            message=msg,
                            wait_task=True
                        )
                    )
                    self.__message_queue.pop(0)

            await self.__start_consume()

            while not connection.is_closed and not self.__is_terminated:
                await asyncio.sleep(1)

            if self.__if_log:
                self.__log(f"Connection is closed")

    async def __run_async(self, loop: AbstractEventLoop | None):
        self.__middleware_chain = self.__build_middleware_chain(
            middlewares=self.__middlewares,
            final_handler=self.__process_message
        )
        while self.__is_terminated is False:
            try:
                await self.__start_connection(loop=loop)
            except Exception as e:
                if self.__if_log:
                    self.__log(f"Exception:{e}")
            await asyncio.sleep(2)
            if self.__if_log:
                self.__log("Reconnecting")

    def task(
            self,
            name: str,
            log_body: bool = True
    ):
        def func_decorator(func):
            self.__tasks[name] = MqTaskRegister(
                name=name,
                func=func,
                log_body=log_body
            )
            return func

        return func_decorator

    def middleware(
            self,
            middleware: Callable[[MqMessageContext, Callable[[MqMessageContext], Any]], Any]
    ):
        self.__middlewares.append(middleware)

    @property
    def loop(self):
        return self.__loop

    def run(self, event_loop: AbstractEventLoop | None = None):
        self.__loop = event_loop or asyncio.get_event_loop()
        self.__loop.run_until_complete(self.__run_async(self.__loop))
        self.__loop.run_forever()

    async def run_async(self, event_loop: AbstractEventLoop | None = None):
        self.__loop = event_loop or asyncio.get_event_loop()
        await self.__run_async(loop=self.__loop)

    def terminate(self):
        self.__is_terminated = True



import json
from dataclasses import asdict, is_dataclass
from datetime import datetime
from typing import Any

from pydantic import BaseModel


class JSON:
    @staticmethod
    def __serializer(value: Any):
        if isinstance(value, datetime):
            return value.isoformat()
        return JSON.dumps(value)

    @staticmethod
    def loads(value: str | None):
        return json.loads(value)

    @staticmethod
    def dumps(value: Any) -> str | None:
        data: str | None = None
        if value is not None:
            if isinstance(value, BaseModel):
                bm: BaseModel = value
                data = bm.json()
            elif is_dataclass(value):
                data = json.dumps(asdict(value), default=JSON.__serializer)
            else:
                data = json.dumps(value, default=JSON.__serializer)
        return data


def is_valid_replay_topic(topic: str | None) -> bool:
    return topic is not None and topic is str and len(topic.strip()) > 0


def to_json_bytes(body: bytes | str | object | None = None) -> bytes | None:
    if body is None:
        return None

    data: bytes
    if body is not None:
        if isinstance(body, bytes):
            data = body
        elif isinstance(body, str):
            data = body.encode()
        else:
            data = JSON.dumps(body).encode()
    else:
        data = bytes()
    return data


def deprecated(
        msg: str | None = None
):
    print(f"[WARNING] deprecated: {msg}")

    def func_decorator(func):
        print(f"[WARNING] deprecated {func.__name__} {msg or ''}")
        return func

    return func_decorator



from asyncio import AbstractEventLoop
from logging import Logger

from mqtasks.body import MqTaskBody
from mqtasks.response_status import MqResponseStatus


class MqTaskMessage:
    loop: AbstractEventLoop
    logger: Logger

    message_id: str
    name: str
    id: str
    body: MqTaskBody
    status: MqResponseStatus

    def __init__(
            self,
            logger: Logger,
            loop: AbstractEventLoop,
            message_id: str,
            task_name: str,
            task_id: str,
            task_body: MqTaskBody,
            status: MqResponseStatus,
    ):
        self.logger = logger
        self.loop = loop
        self.message_id = message_id
        self.name = task_name
        self.id = task_id
        self.body = task_body
        self.status = status

    def __str__(self):
        return f"({MqTaskMessage.__name__} task_id:{self.id} task_name:{self.name} message_id:{self.message_id} body:{self.body.body})"
